import importlib
import os
import subprocess

from git import Repo

from autopack.api import PackResponse, get_pack_data
from autopack.errors import AutoPackError
from autopack.get_pack import get_pack, try_get_pack


def is_dependency_installed(dependency: str) -> bool:
    try:
        importlib.import_module(dependency)
        return True
    except ImportError:
        return False


def install_dependency(dependency: str):
    try:
        subprocess.check_output(
            ["pip", "install", dependency],
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        print(f"{dependency} has been successfully installed.")
    except subprocess.CalledProcessError as e:
        print(f"Installation of {dependency} failed with the following error:")
        print(e.output)


def ask_to_install_dependencies(dependencies: list[str], force=False):
    for dependency in dependencies:
        if is_dependency_installed(dependency):
            continue

        if force:
            install_dependency(dependency)
        else:
            print(
                f"This pack requires the dependency {dependency} to be installed. Continue?"
            )
            agree = input("[Yn]")
            if agree.lower() == "y" or agree == "":
                install_dependency(dependency)
            else:
                print(f"Skipping install of {dependency}")


def install_from_git(pack_data: PackResponse):
    os.makedirs(".autopack", exist_ok=True)

    url = f"https://github.com/{pack_data.author}/{pack_data.repository}.git"
    pack_path = os.path.join(".autopack", pack_data.pack_path())

    if os.path.exists(pack_path):
        print("Repo already exists, pulling updates")
        Repo(pack_path).remotes.origin.pull()
    else:
        print(f"Cloning repo into {pack_path}")
        Repo.clone_from(url, pack_path)


def install_pack(pack_id, force_dependencies=False):
    print(f"Installing pack: {pack_id}")

    pack = try_get_pack(pack_id)
    if pack:
        print(f"Pack {pack_id} already installed.")
        return True

    try:
        pack_data = get_pack_data(pack_id)

        ask_to_install_dependencies(pack_data.dependencies, force_dependencies)
    except AutoPackError as e:
        # Maybe do something else
        print(f"Could not install pack {e}")
        return False
    except BaseException as e:
        print(f"Could not install pack {e}")
        return False

    if pack_data.source == "git":
        install_from_git(pack_data)

    pack = get_pack(pack_id)
    if pack:
        return True

    print("Installation completed but pack could still not be found.")
    return False



import os
from dataclasses import dataclass
from typing import Any, Union
from urllib.parse import urljoin

import requests
from dataclasses_json import dataclass_json
from marshmallow import ValidationError

from autopack.errors import AutoPackFetchError

API_URL = os.environ.get("API_URL", "https://autopack.ai")


@dataclass_json
@dataclass
class PackResponse:
    pack_id: str
    author: str
    repository: str
    module_path: str
    description: str
    name: str
    dependencies: list[str]
    source: str
    arguments: dict[str, Any]

    def pack_path(self) -> str:
        return f"{self.author}-{self.repository}-{self.name}"


def get_pack_data(pack_id: str) -> Union[PackResponse, None]:
    endpoint = "/packs/get"
    url = urljoin(API_URL, endpoint)
    params = {"id": pack_id}

    response = requests.get(url, params=params)
    if response.status_code == 200:
        data = response.json()

        try:
            return PackResponse(**data)
        except (ValidationError, TypeError) as e:
            message = f"Pack fetch received invalid data: {e}"
            print(message)
            raise AutoPackFetchError(message)

    elif response.status_code <= 500:
        print(f"Error: {response.status_code}")
        return None
    else:
        print(f"Error: {response.status_code}")
        error_message = f"Error: {response.status_code}"
        raise AutoPackFetchError(error_message)



import importlib
import inspect
import os
import sys
from typing import Type, Union

from autopack.api import PackResponse, get_pack_data
from autopack.errors import (AutoPackError, AutoPackLoadError,
                             AutoPackNotFoundError, AutoPackNotInstalledError)


def try_get_pack(pack_id: str) -> Union[Type, None]:
    """
    Get a pack based on its ID, in `author/repo_name/pack_name` format. Same as `get_pack` but does not raise an
    Exception. If there is a problem finding or loading a pack it will return None.

    Args:
        pack_id (str): The ID of the pack to fetch.

    Returns:
        BaseTool or None: The fetched pack as a LangChain BaseTool object, None if the pack could not be loaded
    """

    try:
        return get_pack(pack_id)
    except AutoPackError:
        return None


def get_pack(pack_id: str) -> Type:
    """
    Get a pack based on its ID, in `author/repo_name/pack_name` format.

    Args:
        pack_id (str): The ID of the pack to fetch.

    Returns:
        BaseTool: The fetched pack as a LangChain BaseTool object

    Raises:
        AutoPackFetchError: If there was an error during the data fetch.
        AutoPackNotInstalledError: If the pack was found but not installed.
        AutoPackNotFoundError: If no pack matching that ID was found.
        AutoPackLoadError: If the pack was found but there was an error importing or finding the pack class.
    """
    pack_data = get_pack_data(pack_id)

    if not pack_data:
        raise AutoPackNotFoundError()

    return find_pack(pack_data)


def find_module(module_path: str):
    try:
        return importlib.import_module(module_path)
    except ModuleNotFoundError:
        find_autopack_dir_and_add_to_sys_path()
        return importlib.import_module(module_path)


def find_autopack_dir_and_add_to_sys_path(depth=0):
    """Search in current and parent directories looking for .autopack"""
    autopack_dir = os.path.abspath(
        os.path.join(os.getcwd(), *[os.pardir] * depth, ".autopack")
    )

    if not os.path.exists(autopack_dir) or not os.path.isdir(autopack_dir):
        if depth > 3:
            raise ModuleNotFoundError(".autopack directory could not be found")
        return find_autopack_dir_and_add_to_sys_path(depth=depth + 1)

    sys.path.append(autopack_dir)


def find_pack(pack_data: PackResponse) -> Type:
    if pack_data.source == "git":
        module_path = f"{pack_data.pack_path()}.{pack_data.module_path}"
    else:
        module_path = pack_data.module_path

    try:
        module = find_module(module_path)
        if not module:
            message = (
                f"Pack {pack_data.pack_id} could not be found. Either it is misconfigured or .autopack directory not "
                f"found"
            )
            raise AutoPackNotFoundError(message)

        for _, obj in inspect.getmembers(module):
            if is_valid_pack(obj, pack_data.name):
                return obj

        message = (
            f"Pack {pack_data.pack_id} found, but {pack_data.name} is not found in its module, "
            f"{module_path}"
        )
        raise AutoPackNotFoundError(message)
    except ModuleNotFoundError:
        message = f"Pack {pack_data.pack_id} is available but not installed. To install: autopack install {pack_data.pack_id}"
        print(message)
        raise AutoPackNotInstalledError(message)
    except (ImportError, AttributeError) as e:
        message = f"Error loading {pack_data.pack_id}: {e}"
        print(message)
        raise AutoPackLoadError(message)


def is_valid_pack(klass: Type, name: str):
    if not inspect.isclass(klass):
        return False

    base_class_names = [k.__name__ for k in klass.__bases__]
    roughly_adheres_to_interface = (
        hasattr(klass, "run") or "BaseTool" in base_class_names
    )
    if not roughly_adheres_to_interface:
        return False

    # Pack name is the class name
    if name == klass.__name__:
        return True

    # Pack class has a name class variable
    if hasattr(klass, "name"):
        if callable(klass.name):
            klass_name = klass.name()
        else:
            klass_name = klass.name
        return klass_name == name

    # Pack class has a __fields__ class variable (e.g. LangChain's BaseTool)
    if hasattr(klass, "__fields__"):
        name_field = klass.__fields__.get("name")
        return name_field and name_field.default == name

    return False
